from collections import deque
from collections.abc import Iterable
from typing import override

from procset import ProcSet

from pybatsim.batsim import (
    Batsim,
    EDCHelloEvent,
    ExecuteJobEvent,
    Job,
    JobCompletedEvent,
    JobSubmittedEvent,
    RejectJobEvent,
    RxEvent,
    Scheduler,
    SimulationBeginsEvent,
    SimulationEndsEvent,
    SimulationFeatures,
)


class FCFSScheduler(Scheduler):
    waiting_queue: deque[Job]
    idle_resources: ProcSet | None = None
    cluster_size: int | None = None

    @override
    def __init__(self, batsim: Batsim, options: str | None = None):
        super().__init__(batsim, options)

        self.batsim.simulation_metadata.requested_features |= (
            SimulationFeatures.FORWARD_PROFILES_ON_JOB_SUBMISSION
        )
        edc_hello_event = EDCHelloEvent(
            timestamp=self.batsim.time,
            simulation_metadata=self.batsim.simulation_metadata,
            edc_name=type(self).__name__,
            edc_version='0.0.0',
            edc_commit='',
        )
        self.batsim.append_event(edc_hello_event)
        self.waiting_queue = deque()

    @override
    def handle_msg(self, msg: Iterable[RxEvent]) -> None:
        super().handle_msg(msg)
        self.schedule_jobs()

    @override
    def handle_simulation_begins(self, event: SimulationBeginsEvent) -> None:
        self.cluster_size = event.computation_host_number
        self.idle_resources = ProcSet((0, self.cluster_size - 1))

    @override
    def handle_simulation_ends(self, event: SimulationEndsEvent) -> None:
        pass

    @override
    def handle_submitted_job(self, event: JobSubmittedEvent) -> None:
        assert self.cluster_size is not None, (
            'expected initialized cluster_size upon JobSubmittedEvent'
        )

        job = event.job

        if job.resource_request > self.cluster_size:
            # Job requests more resources than cluster_size: trivial reject.
            reject_event = RejectJobEvent(timestamp=self.batsim.time, job=job)
            self.batsim.append_event(reject_event)
        else:
            self.waiting_queue.append(job)

    @override
    def handle_completed_job(self, event: JobCompletedEvent) -> None:
        assert self.idle_resources is not None, (
            'expected initialized idle_resources upon JobCompletedEvent'
        )
        self.idle_resources |= event.job.allocation

    def schedule_jobs(self) -> None:
        scheduled_jobs: list[Job] = []

        while self.waiting_queue:
            # Stop scheduling jobs on the first job lacking resources.
            if self.waiting_queue[0].resource_request > len(self.idle_resources):
                break

            # Extract job from waiting queue.
            job = self.waiting_queue.popleft()

            # Allocate resources to job.
            resources = self.idle_resources[: job.resource_request]
            job.allocation = ProcSet(*resources)
            self.idle_resources -= job.allocation

            # Add job to the pending schedule batch.
            scheduled_jobs.append(job)

        # Send the list of scheduled jobs to Batsim.
        now = self.batsim.time
        for job in scheduled_jobs:
            execute_event = ExecuteJobEvent(timestamp=now, job=job)
            self.batsim.append_event(execute_event)



"""
Main simulation components: batsim proxy and external decision component interfaces.
"""

from __future__ import annotations

import json
import sys
from abc import abstractmethod
from collections import deque
from collections.abc import Iterable
from types import TracebackType
from typing import Literal, Protocol, Self, cast

import zmq

from .batprotocol import Duration, EventPDict, JobId, MessagePDict, Timestamp
from .core import SimulationFeatures, SimulationMetadata
from .events import (
    AllStaticExternalEventsHaveBeenInjectedEvent,
    AllStaticJobsHaveBeenSubmittedEvent,
    EDCHelloEvent,
    ExternalEventOccurredEvent,
    HostsPStateChangedEvent,
    HostsTurnedOnOffEvent,
    JobCompletedEvent,
    JobsKilledEvent,
    JobSubmittedEvent,
    KillJobsEvent,
    RegisterJobEvent,
    RejectJobEvent,
    RequestedCallEvent,
    RxEvent,
    SimulationBeginsEvent,
    SimulationEndsEvent,
    TxEvent,
)
from .job import Job

# SERIALIZATION_FORMAT_BINARY = 1  # unsupported
SERIALIZATION_FORMAT_JSON = 2

WORKLOAD_JOB_SEPARATOR = '!'
# ATTEMPT_JOB_SEPARATOR = '#'  # unsupported, used when resubmitting jobs


# TODO: Implement public API to access SimulationMetadata
class Batsim:
    # Communication with the batsim process.
    _endpoint: str
    _timeout: int
    _zmq_socket: zmq.Socket[bytes] | None

    # Decision logic, injected by the user.
    _edc: ExternalDecisionComponent | None

    # Simulation-related attributes.
    _time: Timestamp
    _simulation_metadata: SimulationMetadata
    _rx: deque[RxEvent]
    _tx: deque[TxEvent]
    _alive_jobs: dict[JobId, Job]
    _zombie_jobs: dict[JobId, Job]

    # XXX: This is fragile, we should rather track the state from the state-machine.
    _received_SimulationEndsEvent: bool  # noqa: N815 (reason: use event name)

    def __init__(self, *, endpoint: str, timeout: int | None = None):
        self._endpoint = endpoint
        self._timeout = -1 if timeout is None else timeout
        self._zmq_socket = None

        self._edc = None

        self._time = 0.0
        self._simulation_metadata = SimulationMetadata()
        self._rx = deque()
        self._tx = deque()

        self._received_SimulationEndsEvent = False

        # Batsim protocol can either send a job id or a full job description.
        # The Python API abstracts this away, and only works with Job.
        # _alive_jobs and _zombie_jobs ensure each Job object lives at least as
        # long as needed: an EDC may further extend a job's life.
        # _zombie_jobs extends the life of jobs with a pending death acknowledgment.
        #
        # The state machine of jobs, as known by pybatsim, is summarized below.
        # Initial states are drawn as rectangle with single lines.
        # Final states are drawn as rectangles with double lines.
        # Other states are drawn as dashed ovals.
        #
        # States are marked with A for jobs in _alive_jobs, Z for jobs in _zombie_jobs.
        # In the 'declared' state, only dynamic jobs are known by pybatsim:
        # static jobs are created when Batsim sends a JobSubmittedEvent.
        #
        # ┌────────────┐  ┌─────────────┐
        # │ static job │  │ dynamic job │───────────┐
        # └────────────┘  └─────────────┘           │
        #        │               │                  │
        #        │            ack &&           not(ack) &&
        #        │        RegisterJob[tx]    RegisterJob[tx]
        #        │               🠇                  │
        #        │         ╭╌╌╌╌╌╌╌╌╌╌╮             │
        #        └──╴ε╶───>╎ declared ╎             │
        #                  ╎     A*   ╎             │
        #                  ╰╌╌╌╌╌╌╌╌╌╌╯             │
        #                        │                  │
        #                 JobSubmitted[rx]          │
        #                        🠇                  │
        #                  ╭╌╌╌╌╌╌╌╌╌╌╌╮            │
        #       ┌──────────╎ submitted ╎<───────────┘
        #       │          ╎     A     ╎
        #       │          ╰╌╌╌╌╌╌╌╌╌╌╌╯
        #       │                │
        # RejectJob[tx]    ExecuteJob[tx]
        #       🠇                🠇
        # ╔══════════╗      ╭╌╌╌╌╌╌╌╌╌╮                      ╭╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╮
        # ║ rejected ║      ╎ running ╎────╴KillJob[tx]╶────>╎ running,kill_sent ╎
        # ╚══════════╝      ╎    A    ╎                      ╎        A,Z        ╎
        #                   ╰╌╌╌╌╌╌╌╌╌╯                      ╰╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╯
        #                        │                                     │
        #                 JobCompleted[rx]                      JobCompleted[rx]
        #                        🠇                                     🠇
        #                 ╔═════════════╗                  ╭╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╮
        #                 ║ completed_* ║──╴KillJob[tx]╶──>╎ completed_*,kill_sent ╎
        #                 ╚═════════════╝                  ╎           Z           ╎
        #                                                  ╰╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╯
        #                                                              │
        #                                                       JobKillled[rx]
        #                                                              🠇
        #                                                 ╔════════════════════════╗
        #                                                 ║ completed_*,kill_acked ║
        #                                                 ╚════════════════════════╝
        self._alive_jobs = {}
        self._zombie_jobs = {}

    def __setup_zmq(self) -> None:
        context = zmq.Context()
        context.setsockopt(zmq.RCVTIMEO, self._timeout)

        self._zmq_socket = context.socket(socket_type=zmq.REP)
        self._zmq_socket.bind(self._endpoint)
        # Replace _endpoint with the actual endpoint in use (no wildcard): this
        # is needed for the call to unbind() during teardown.
        self._endpoint = cast(str, self._zmq_socket.last_endpoint)

    def __teardown_zmq(self) -> None:
        assert self._zmq_socket is not None, 'uninitialized _zmq_socket'
        self._zmq_socket.unbind(self._endpoint)
        self._zmq_socket.close()
        self._zmq_socket.context.destroy()

    def __enter__(self) -> Self:
        self.__setup_zmq()  # connect to Batsim ØMQ socket

        try:
            # handle init message
            self._recv_init_msg()

        except Exception:  # XXX: consider reducing caught exceptions
            self.__teardown_zmq()
            raise

        else:
            return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> Literal[False]:
        self.__teardown_zmq()  # always clean the Batsim ØMQ resources (context/socket)
        return False  # always propagate exceptions occurring in the with block

    @property
    def connected(self) -> bool:
        return self._zmq_socket is not None and not self._zmq_socket.closed

    @property
    def time(self) -> Timestamp:
        return self._time

    def consume_time(self, duration: Duration) -> None:
        assert duration > 0, 'cannot go back in time'
        self._time += duration

    @property
    def simulation_metadata(self) -> SimulationMetadata:
        return self._simulation_metadata

    def _recv_init_msg(self) -> None:
        # Batsim sends the first message (init message)
        # message format:
        # 1. init_data_size(uint32)
        # 2. init_data(init_data_size bytes):
        #    EDC initialization string forwarded from Batsim CLI
        assert self._zmq_socket is not None, 'uninitialized _zmq_socket'

        raw_msg: bytes = self._zmq_socket.recv()

        data_size = int.from_bytes(raw_msg[:4], byteorder=sys.byteorder)
        data: bytes = raw_msg[4:]
        if len(raw_msg) != 4 + data_size:
            err_msg = (
                'invalid data_size: '
                f'received init message is {len(raw_msg)} bytes long, '
                f'read data is {data_size} bytes long (should be 4 bytes less)'
            )
            raise ValueError(err_msg)

        self._simulation_metadata.edc_init_str = data[:data_size].decode('utf-8')

    def _send_edc_hello_msg(self) -> None:
        # Prepare and send the first EDC message (answer to the init message)
        # message format:
        # 1. serialization_format(uint32)
        # 2. serialized message with the EDCHelloEvent
        assert self._zmq_socket is not None, 'uninitialized _zmq_socket'

        # Check _tx buffer contains a single event of type EDCHelloEvent.
        if not isinstance(self._tx[0], EDCHelloEvent):
            err_msg = (
                f"EDC asked to send '{type(self._tx[0]).__name__}', "
                "expected 'EDCHelloEvent'"
            )
            raise TypeError(err_msg)
        if len(self._tx) != 1:
            err_msg = "EDC hello message must contain a single 'EDCHelloEvent'"
            raise ValueError(err_msg)

        # Build sequence of bytes to send.
        serialization_format: bytes = SERIALIZATION_FORMAT_JSON.to_bytes(
            4, byteorder='little'
        )
        msg_dict: MessagePDict = self.serialize_msg()
        raw_msg: str = json.dumps(msg_dict)
        wire_msg: bytes = serialization_format + raw_msg.encode()

        self._zmq_socket.send(wire_msg)
        self._tx.clear()

    def register_edc(self, edc: ExternalDecisionComponent) -> None:
        self._edc = edc
        # XXX:
        #   This sequence is fragile as it requires:
        #   1. the EDC to append the hello event during its initialization
        #   2. no other creation of messages before registering
        #
        #   Consider introducing a required method craft_hello_event on
        #   ExternalDecisionComponent.
        self._send_edc_hello_msg()

    def is_simulation_finished(self) -> bool:
        return self._received_SimulationEndsEvent

    def recv_msg(self) -> None:
        assert self._zmq_socket is not None, 'uninitialized _zmq_socket'

        try:
            raw_msg = self._zmq_socket.recv_string()
            # XXX: https://framagit.org/batsim/batprotocol/-/issues/3
            #   The batprotocol appends a null byte after each sent JSON message.
            #   Consider removing the null byte from the protocol as ØMQ handles
            #   the length of sent messages.
            #   This would allow to use self._zmq_socket.recv_json()
            raw_msg = raw_msg[:-1]  # drop terminating null byte
            protocol_dict = cast(MessagePDict, json.loads(raw_msg))
            print(f'Received Batsim message: {protocol_dict}')
            self.deserialize_msg(protocol_dict)

        except Exception:  # noqa: TRY203 (reason: due to the TODO on the next line)
            # TODO: handle json.loads and deserialization exceptions
            raise

    def dispatch_msg(self) -> None:
        """Trigger handling of the received message by the registered EDC."""
        assert self._edc is not None, 'uninitialized _edc'
        self._edc.handle_msg(self._rx)

    def send_msg(self) -> None:
        """Send the built answer message to Batsim."""
        assert self._zmq_socket is not None, 'uninitialized _zmq_socket'
        protocol_dict = self.serialize_msg()
        print(f'Sending to Batsim:\n{protocol_dict}')
        self._zmq_socket.send_json(protocol_dict)
        self._tx.clear()

    def pop_event(self) -> RxEvent:
        return self._rx.popleft()

    def append_event(self, event: TxEvent) -> None:
        """Add event to the next message for Batsim."""
        if isinstance(event, RejectJobEvent):
            # Remove from _alive_jobs as RejectJobEvent is the last possible
            # event sent to Batsim.
            assert event.job.id_ not in self._zombie_jobs
            del self._alive_jobs[event.job.id_]

        elif isinstance(event, KillJobsEvent):
            # Extend lifetime of of jobs until we receive their death acknowledgment.
            for job in event.jobs:
                self._zombie_jobs[job.id_] = job

        elif isinstance(event, RegisterJobEvent):
            # event.job is a dynamic job created by the EDC: keep track of it.
            assert event.job.id_ not in self._alive_jobs
            self._alive_jobs[event.job.id_] = event.job

        self._tx.append(event)

    def deserialize_event(self, protocol_dict: EventPDict) -> RxEvent:
        # For events only containing a job id, retrieve the corresponding Job
        # object from _alive_jobs and inject it in protocol_dict.
        # This allows RxEvent.from_protocol_dict to work with the correct objet.
        if protocol_dict['event_type'] == 'JobCompletedEvent':
            # Remove job from _alive_jobs as this is the last related event
            # received from Batsim.
            # The job may still be present in _zombie_jobs.
            job = self._alive_jobs.pop(protocol_dict['event']['job_id'])
            protocol_dict['__pybatsim_job'] = job

        elif protocol_dict['event_type'] == 'JobsKilledEvent':
            dead_jobs = []
            for job_id in protocol_dict['event']['job_ids']:
                assert job_id not in self._alive_jobs
                job = self._zombie_jobs.pop(job_id)
                dead_jobs.append(job)
            protocol_dict['__pybatsim_dead_jobs'] = dead_jobs

        event = RxEvent.from_protocol_dict(protocol_dict)

        if isinstance(event, SimulationEndsEvent):
            self._received_SimulationEndsEvent = True

        elif isinstance(event, JobSubmittedEvent):
            # Batsim sends a full job description in a JobSubmittedEvent.
            # The described job is either static or dynamic.
            #   - in the former case, this is a new job: store it in _alive_jobs.
            #   - in the latter case, retrieve the existing job and update in place.
            new_job = event.job
            job = self._alive_jobs.setdefault(new_job.id_, new_job)

            if job is not new_job:
                # Dynamic jobs are submitted back by Batsim only if an
                # acknowledgment is requested.
                assert (
                    SimulationFeatures.ACKNOWLEDGE_DYNAMIC_JOBS
                    in self.simulation_metadata.requested_features
                )

                # Update the existing dynamic job in place.
                assert job.submission_time is None, 'overwriting job.submission_time'
                job.submission_time = new_job.submission_time

                # Reuse the existing dynamic job in the deserialized event.
                event.job = job

        return event

    def deserialize_msg(self, protocol_dict: MessagePDict) -> None:
        self._rx.clear()  # drop previous msg

        assert self._time <= protocol_dict['now'], 'decreasing simulation time'
        self._time = protocol_dict['now']

        # Fill _rx buffer with the events received in current msg.
        for event_dict in protocol_dict['events']:
            print('--- Received event of type', event_dict['event_type'])
            event = self.deserialize_event(event_dict)
            self._rx.append(event)

    def serialize_msg(self) -> MessagePDict:
        return {
            'now': self._time,
            'events': [event.to_protocol_dict() for event in self._tx],
        }


class ExternalDecisionComponent(Protocol):
    def __init__(self, batsim: Batsim, options: str | None = None): ...

    def handle_msg(self, msg: Iterable[RxEvent]) -> None: ...

    def finalize(self) -> None: ...


class Scheduler(ExternalDecisionComponent):
    batsim: Batsim
    _options: str | None

    def __init__(self, batsim: Batsim, options: str | None = None):
        self.batsim = batsim
        self._options = options

    def handle_msg(self, msg: Iterable[RxEvent]) -> None:
        for event in msg:
            self._dispatch(event)

    def _dispatch(self, event: RxEvent) -> None:
        match event:
            case JobSubmittedEvent():
                self.handle_submitted_job(event)
            case JobCompletedEvent():
                self.handle_completed_job(event)
            case JobsKilledEvent():
                self.handle_jobs_killed(event)
            case RequestedCallEvent():
                self.handle_requested_call(event)
            case ExternalEventOccurredEvent():
                self.handle_external_event_occurred(event)
            case HostsPStateChangedEvent():
                self.handle_hosts_pstate_changed(event)
            case HostsTurnedOnOffEvent():
                self.handle_hosts_turned_onoff(event)
            case SimulationBeginsEvent():
                self.handle_simulation_begins(event)
            case SimulationEndsEvent():
                self.handle_simulation_ends(event)
            case AllStaticJobsHaveBeenSubmittedEvent():
                self.handle_no_more_static_jobs(event)
            case AllStaticExternalEventsHaveBeenInjectedEvent():
                self.handle_no_more_external_events(event)
            case _:
                err_msg = f"Unknown Event '{type(event).__name__}'"
                raise TypeError(err_msg)

    @abstractmethod
    def handle_simulation_begins(self, event: SimulationBeginsEvent) -> None: ...

    @abstractmethod
    def handle_simulation_ends(self, event: SimulationEndsEvent) -> None: ...

    @abstractmethod
    def handle_submitted_job(self, event: JobSubmittedEvent) -> None: ...

    @abstractmethod
    def handle_completed_job(self, event: JobCompletedEvent) -> None: ...

    def handle_jobs_killed(self, event: JobsKilledEvent) -> None:
        pass

    def handle_external_event_occurred(self, event: ExternalEventOccurredEvent) -> None:
        pass

    def handle_hosts_pstate_changed(self, event: HostsPStateChangedEvent) -> None:
        pass

    def handle_hosts_turned_onoff(self, event: HostsTurnedOnOffEvent) -> None:
        pass

    def handle_requested_call(self, event: RequestedCallEvent) -> None:
        pass

    def handle_no_more_static_jobs(
        self,
        event: AllStaticJobsHaveBeenSubmittedEvent,
    ) -> None:
        pass

    def handle_no_more_external_events(
        self,
        event: AllStaticExternalEventsHaveBeenInjectedEvent,
    ) -> None:
        pass

    def finalize(self) -> None:
        pass



"""
pybatsim.cmdline
~~~~~~~~~~~~~~~~

Command line interface.
"""

import argparse
import collections
import inspect
import io
import json
import logging
import textwrap

from . import __version__
from .batsim import Batsim, ExternalDecisionComponent
from .plugin import (
    EDC_ENTRY_POINT,
    find_ambiguous_edc_names,
    find_plugin_edcs,
)


class _JsonStoreAction(argparse.Action):
    """
    Decode and store the JSON-encoded value of a single argument.

    If the argument's value starts with a '@', the path to a JSON file is
    expected.
    Otherwise, a valid JSON string is expected.
    """

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            err_msg = 'nargs is not allowed'
            raise ValueError(err_msg)
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, _parser, namespace, values, _option_string=None):
        try:
            rawcontent = values.strip()

            if rawcontent.startswith('@'):
                # classic text stream of the file containing the options
                json_file = open(rawcontent[1:], encoding='utf-8')  # noqa: SIM115 (reason: false positive)
            else:
                # encapsulate the whole JSON string in a text stream
                json_file = io.StringIO(rawcontent)

            with json_file:
                decoded_content = json.load(json_file)
                setattr(namespace, self.dest, decoded_content)

        except OSError as err:
            # raised by open()
            raise argparse.ArgumentError(
                self,
                f"unable to read '{err.filename}': {err.strerror.lower()}",
            ) from None
        except json.JSONDecodeError:
            # raised by json.load(), subclass of ValueError
            raise argparse.ArgumentError(self, 'invalid JSON object') from None
        except ValueError:
            # raised by open() or json.load()
            raise argparse.ArgumentError(
                self,
                'incorrect encoding (expected utf-8)',
            ) from None


class _ListExternalDecisionComponentsAction(argparse.Action):
    def __init__(
        self,
        option_strings,
        dest=argparse.SUPPRESS,
        default=argparse.SUPPRESS,
        help='list known External Decision Components (EDCs) and exit',  # noqa: A002
    ):
        super().__init__(option_strings, dest, default=default, nargs=0, help=help)

    def __call__(self, parser, _namespace, _values, _option_string=None):
        # organize names by actual EDC class (some names can be aliases)
        known_edcs_by_class = collections.defaultdict(list)
        for name, cls in find_plugin_edcs():
            known_edcs_by_class[cls].append(name)
        # display names of EDC in alphabetical order
        for names in known_edcs_by_class.values():
            names.sort()
        for cls, names in known_edcs_by_class.items():
            doc = cls.__doc__
            doc = (
                inspect.cleandoc(doc).splitlines()[0]
                if doc is not None
                else cls.__qualname__
            )
            print(', '.join(names) + ':')
            print('  ' + doc)
        parser.exit()


def _build_parser():
    parser = argparse.ArgumentParser(
        description='Run a PyBatsim External Decision Component (EDC).',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent("""\
            exit status:
              %(prog)s can exit with the following return codes:
                0  success
                1  simulation failure
                2  argument parsing error
        """),
    )
    parser.add_argument(
        '--version',
        action='version',
        version=__version__,
    )
    parser.add_argument(
        '--list-external-decision-components',
        '--list-edcs',
        action=_ListExternalDecisionComponentsAction,
    )
    parser.add_argument(
        '-t',
        '--timeout',
        default=5_000,
        type=int,
        help=(
            'the timeout (in milliseconds) to wait for a Batsim answer, '
            'supply a negative value to disable '
            '(default: 5000)'
        ),
    )
    parser.add_argument(
        '-s',
        '--socket-endpoint',
        default='tcp://*:28000',
        help=(
            'address of Batsim socket, '
            "formatted as 'protocol://interface:port' "
            '(default: tcp://*:28000)'
        ),
        metavar='ADDRESS',
    )
    parser.add_argument(
        '-o',
        '--edc-options',
        default={},
        action=_JsonStoreAction,
        help=(
            'options forwarded to the External Decision Component '
            '(default: empty dict), '
            'either a JSON string (e.g., \'{"option": "value"}\') '
            "or a @-prefixed JSON file containing the options (e.g., '@options.json')"
        ),
        metavar='[@]OPTIONS',
    )
    parser.add_argument(
        'edc_name',
        choices=sorted({name for name, _ in find_plugin_edcs()}),
        metavar='EDC',
        help=(
            'name of the External Decision Component (EDC) to run, '
            f"as registered under '{EDC_ENTRY_POINT}' entry point"
        ),
    )
    return parser


def _abort_on_ambiguous_edc_name(name, *, parser) -> None:
    ambiguous_names = find_ambiguous_edc_names()
    if name in ambiguous_names:
        err_msg = (
            f"overlapping bindings in '{EDC_ENTRY_POINT}' entry point, "
            'check your packaging! '
            f"'{name}' is defined more than once, and binds to: "
        )
        err_msg += ', '.join(ambiguous_names[name])
        parser.error(err_msg)


def _lookup_edc_class(name) -> type[ExternalDecisionComponent]:
    """Lookup an EDC by name."""
    for found_name, cls in find_plugin_edcs():
        if name == found_name:
            return cls

    err_msg = f'unknown EDC name: {name}'
    raise LookupError(err_msg)


def main(args=None) -> None:
    logging.basicConfig(level=logging.INFO)

    # retrieve arguments
    parser = _build_parser()
    arguments = parser.parse_args(args)
    logging.debug(f'parsed arguments: {vars(arguments)}')

    # retrieve class of requested EDC
    _abort_on_ambiguous_edc_name(arguments.edc_name, parser=parser)
    edc_cls = _lookup_edc_class(arguments.edc_name)

    # TODO: handle exceptions from ØMQ
    # A ØMQ timeout usually means batsim has deadlocked/crashed
    with Batsim(
        endpoint=arguments.socket_endpoint, timeout=arguments.timeout
    ) as batsim:
        edc: ExternalDecisionComponent = edc_cls(batsim, options=arguments.edc_options)

        batsim.register_edc(edc)

        while not batsim.is_simulation_finished():
            batsim.recv_msg()
            batsim.dispatch_msg()
            batsim.send_msg()

        edc.finalize()



"""
pybatsim.plugin
~~~~~~~~~~~~~~~

PyBatsim plugin interface.

Register user-defined External Decision Component (EDC) via the entry point mechanism.

The following snippet shows how to register under the name
``yourschedulername`` the user-defined EDC :py:class:`YourScheduler`
defined in the module :py:mod:`yourscheduler`.

.. code-block:: cfg

   [pybatsim.external_decision_components]
   yourschedulername = yourscheduler:YourScheduler


Refer to the documentation of entry points of your packaging tool to
register an EDC.
"""

import collections
from importlib.metadata import entry_points

EDC_ENTRY_POINT = 'pybatsim.external_decision_components'


def find_plugin_edcs():
    """Yield the tuples (name, class) for known External Decision Component (EDC)."""
    for edc in entry_points(group=EDC_ENTRY_POINT):
        yield edc.name, edc.load()


def find_ambiguous_edc_names():
    """
    Return the dict of names bound to multiple External Decision Component (EDC).

    For each ambiguous name, the dict maps the name to the set of entry points
    values.
    """
    known_edc_names = collections.defaultdict(set)
    for edc in entry_points(group=EDC_ENTRY_POINT):
        known_edc_names[edc.name].add(edc.value)

    return {
        name: values for (name, values) in known_edc_names.items() if len(values) > 1
    }



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Flag, FlagBoundary, auto
from typing import Final

from .batprotocol import (
    SimulationFeaturesPDict,
    SimulationMetadataPDict,
    __batprotocol_version__,
)


class SimulationFeatures(Flag, boundary=FlagBoundary.STRICT):
    """
    Configuration of simulation features.

    See batprotocol::fb::EDCRequestedSimulationFeatures for reference.
    """

    DYNAMIC_REGISTRATION = auto()
    """Enable dynamic registration of jobs and profiles."""

    PROFILE_REUSE = auto()
    """Allow existing profiles to be used to register new dynamic jobs."""

    ACKNOWLEDGE_DYNAMIC_JOBS = auto()
    """Ask Batsim to emit back a JobSubmittedEvent upon creation of a dynamic job."""

    FORWARD_PROFILES_ON_JOB_SUBMISSION = auto()
    """Include profile information in JobSubmittedEvent."""

    FORWARD_PROFILES_ON_JOBS_KILLED = auto()
    """Include profile information in JobsKilledEvent."""

    FORWARD_PROFILES_ON_SIMULATION_BEGINS = auto()
    """Include profile information in SimulationBeginsEvent."""

    @classmethod
    def default(cls) -> SimulationFeatures:
        return cls(0)

    def to_protocol_dict(self) -> SimulationFeaturesPDict:
        return {
            'dynamic_registration': SimulationFeatures.DYNAMIC_REGISTRATION in self,
            'profile_reuse': SimulationFeatures.PROFILE_REUSE in self,
            'acknowledge_dynamic_jobs': SimulationFeatures.ACKNOWLEDGE_DYNAMIC_JOBS in self,  # noqa: E501 (reason: improves readability)
            'forward_profiles_on_job_submission': SimulationFeatures.FORWARD_PROFILES_ON_JOB_SUBMISSION in self,  # noqa: E501 (reason: improves readability)
            'forward_profiles_on_jobs_killed': SimulationFeatures.FORWARD_PROFILES_ON_JOBS_KILLED in self,  # noqa: E501 (reason: improves readability)
            'forward_profiles_on_simulation_begins': SimulationFeatures.FORWARD_PROFILES_ON_SIMULATION_BEGINS in self,  # noqa: E501 (reason: improves readability)
        }  # fmt: skip


@dataclass
class SimulationMetadata:
    """
    Container for information about the simulation.

    The information contained in SimulationMetadata is set during the initial
    handshake with Batsim.
    """

    edc_init_str: str | None = None

    # Batsim and batprotocol information
    batprotocol_version: Final[str] = __batprotocol_version__
    batsim_version: str | None = None
    batsim_commit: str | None = None

    # simulation features requested by the EDC
    requested_features: SimulationFeatures = field(
        default_factory=SimulationFeatures.default
    )

    def to_protocol_dict(self) -> SimulationMetadataPDict:
        return {
            'batprotocol_version': self.batprotocol_version,
            'requested_simulation_features': self.requested_features.to_protocol_dict(),
        }



"""
Definitions of types describing the structure of the messages exchanged with
batsim.
These descriptions define the pybatsim interpretation of the batprocotol.
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Final,
    Literal,
    LiteralString,
    NotRequired,
    TypedDict,
)

if TYPE_CHECKING:
    from .job import Job

__batprotocol_version__: Final[str] = '1.0.0'


type JobId = str
type ProfileId = str

type HostId = int
type StorageId = str

type CallId = str

type Timestamp = float
type Duration = float


# The wire format of the batprotocol is defined via TypedDict types.
# Keep in mind "a TypedDict is a structural type" and that "TypedDict types […]
# inherit from other TypedDict types to share common items".


# ----- simulation metadata wire format ----


class SimulationFeaturesPDict(TypedDict):
    dynamic_registration: bool
    profile_reuse: bool
    acknowledge_dynamic_jobs: bool
    forward_profiles_on_job_submission: bool
    forward_profiles_on_jobs_killed: bool
    forward_profiles_on_simulation_begins: bool


# XXX(rb): consider merging with related classes for EDCHelloEvent as this is a
# strict subset of the content of the EDCHelloEvent protocol table.
class SimulationMetadataPDict(TypedDict):
    batprotocol_version: str
    batprotocol_commit: NotRequired[str]
    requested_simulation_features: SimulationFeaturesPDict


# ----- events wire format -----


# The Event union of the batprotocol is never directly used: it is always
# wrapped in the EventAndTimestamp table.
# Hence, the _BaseEventPDict describes the structure of the EventAndTimestamp
# table.
class _BaseEventPDict[T: LiteralString, P](TypedDict):
    timestamp: Timestamp
    event_type: T
    event: P


class _EmptyEventPayload(TypedDict):
    pass


class _JobSubmittedEventPayload(TypedDict):
    job_id: JobId
    job: JobPDict
    submission_time: Timestamp
    profile: NotRequired[ProfilePDict]


class _JobCompletedEventPayload(TypedDict):
    job_id: JobId
    state: Literal[
        'COMPLETED_SUCCESSFULLY',
        'COMPLETED_FAILED',
        'COMPLETED_WALLTIME_REACHED',
        'COMPLETED_KILLED',
        'REJECTED',
    ]
    return_code: int


class _RejectJobEventPayload(TypedDict):
    job_id: JobId


class _ExecuteJobEventPayload(TypedDict):
    job_id: JobId
    allocation: JobAllocationPDict
    # profile_allocation_override and storage_placement are not supported yet
    profile_allocation_override: NotRequired[list[ProfileAllocationPDict]]
    storage_placement: NotRequired[list[StorageHostMappingPDict]]


class _KillJobsEventPayload(TypedDict):
    job_ids: list[JobId]


class _JobsKilledEventPayload(TypedDict):
    job_ids: list[JobId]
    progresses: list[Any]  # XXX: replace with semantic type
    profiles: NotRequired[list[ProfilePDict]]


class _RegisterProfileEventPayload(TypedDict):
    profile: ProfilePDict


class _RegisterJobEventPayload(TypedDict):
    job_id: JobId
    job: JobPDict


# CreateProbeEvent is not supported.

# StopProbeEvent is not supported.

# TriggerProbeEvent is not supported.

# ResetProbeEvent is not supported.

# ProbeDataEmittedEvent is not supported.


class _CallMeLaterEventPayload(TypedDict):
    call_me_later_id: CallId
    when_type: Any  # XXX: replace with semantic type
    when: Any  # XXX: replace with semantic type


class _RequestedCallEventPayload(TypedDict):
    call_me_later_id: CallId
    last_periodic_call: bool


class _StopCallMeLaterEventPayload(TypedDict):
    call_me_later_id: CallId


class _EDCHelloEventPayload(TypedDict):
    batprotocol_version: str
    batprotocol_commit: NotRequired[str]
    decision_component_name: str
    decision_component_version: str
    decision_component_commit: NotRequired[str]
    requested_simulation_features: SimulationFeaturesPDict


class _SimulationBeginsEventPayload(TypedDict):
    batprotocol_version: str
    batsim_version: str
    batsim_commit: NotRequired[str]

    host_number: int
    computation_host_number: int
    computation_hosts: list[Any]  # XXX: replace with semantic type
    storage_host_number: int
    storage_hosts: list[Any]  # XXX: replace with semantic type

    batsim_arguments: list[str]
    batsim_execution_context_json: str

    workloads: list[Any]  # XXX: replace with semantic type
    external_event_lists: list[Any]  # XXX: replace with semantic type
    profiles: NotRequired[list[ProfilePDict]]


# SimulationEndsEvent has an empty payload.

# SimulationErrorEvent is not supported.

# AllStaticJobsHaveBeenSubmittedEvent has an empty payload.

# AllStaticExternalEventsHaveBeenInjectedEvent has an empty payload.

# FinishRegistrationEvent has an empty payload.

# ForceSimulationStopEvent has an empty payload.


class _ChangeHostsPStateEventPayload(TypedDict):
    host_ids: str
    pstate: int


class _HostsPStateChangedEventPayload(TypedDict):
    host_ids: str
    pstate: int


class _TurnOnOffHostsEventPayload(TypedDict):
    host_ids: str
    state: int


class _HostsTurnedOnOffEventPayload(TypedDict):
    host_ids: str
    state: int


class _ExternalEventOccurredEventPayload(TypedDict):
    id: str
    external_event_type: str  # XXX: replace with semantic type
    external_event: dict[Any, Any]  # XXX: replace with semantic type


type JobSubmittedEventPDict = _BaseEventPDict[
    Literal['JobSubmittedEvent'],
    _JobSubmittedEventPayload,
]


class JobCompletedEventPDict(
    _BaseEventPDict[
        Literal['JobCompletedEvent'],
        _JobCompletedEventPayload,
    ]
):
    __pybatsim_job: Job


type RejectJobEventPDict = _BaseEventPDict[
    Literal['RejectJobEvent'],
    _RejectJobEventPayload,
]
type ExecuteJobEventPDict = _BaseEventPDict[
    Literal['ExecuteJobEvent'],
    _ExecuteJobEventPayload,
]
type KillJobsEventPDict = _BaseEventPDict[
    Literal['KillJobsEvent'],
    _KillJobsEventPayload,
]


class JobsKilledEventPDict(
    _BaseEventPDict[
        Literal['JobsKilledEvent'],
        _JobsKilledEventPayload,
    ]
):
    __pybatsim_dead_jobs: list[Job]


type RegisterProfileEventPDict = _BaseEventPDict[
    Literal['RegisterProfileEvent'],
    _RegisterProfileEventPayload,
]
type RegisterJobEventPDict = _BaseEventPDict[
    Literal['RegisterJobEvent'],
    _RegisterJobEventPayload,
]
# CreateProbeEventPDict is not supported.
# StopProbeEventPDict is not supported.
# TriggerProbeEventPDict is not supported.
# ResetProbeEventPDict is not supported.
# ProbeDataEmittedEventPDict is not supported.
type CallMeLaterEventPDict = _BaseEventPDict[
    Literal['CallMeLaterEvent'],
    _CallMeLaterEventPayload,
]
type RequestedCallEventPDict = _BaseEventPDict[
    Literal['RequestedCallEvent'],
    _RequestedCallEventPayload,
]
type StopCallMeLaterEventPDict = _BaseEventPDict[
    Literal['StopCallMeLaterEvent'],
    _StopCallMeLaterEventPayload,
]
type EDCHelloEventPDict = _BaseEventPDict[
    Literal['EDCHelloEvent'],
    _EDCHelloEventPayload,
]
type SimulationBeginsEventPDict = _BaseEventPDict[
    Literal['SimulationBeginsEvent'],
    _SimulationBeginsEventPayload,
]
type SimulationEndsEventPDict = _BaseEventPDict[
    Literal['SimulationEndsEvent'],
    _EmptyEventPayload,
]
# SimulationErrorEvent is not supported.
type AllStaticJobsHaveBeenSubmittedEventPDict = _BaseEventPDict[
    Literal['AllStaticJobsHaveBeenSubmittedEvent'],
    _EmptyEventPayload,
]
type AllStaticExternalEventsHaveBeenInjectedEventPDict = _BaseEventPDict[
    Literal['AllStaticExternalEventsHaveBeenInjectedEvent'],
    _EmptyEventPayload,
]
type FinishRegistrationEventPDict = _BaseEventPDict[
    Literal['FinishRegistrationEvent'],
    _EmptyEventPayload,
]
type ForceSimulationStopEventPDict = _BaseEventPDict[
    Literal['ForceSimulationStopEvent'],
    _EmptyEventPayload,
]
type ChangeHostsPStateEventPDict = _BaseEventPDict[
    Literal['ChangeHostsPStateEvent'],
    _ChangeHostsPStateEventPayload,
]
type HostsPStateChangedEventPDict = _BaseEventPDict[
    Literal['HostsPStateChangedEvent'],
    _HostsPStateChangedEventPayload,
]
type TurnOnOffHostsEventPDict = _BaseEventPDict[
    Literal['TurnOnOffHostsEvent'],
    _TurnOnOffHostsEventPayload,
]
type HostsTurnedOnOffEventPDict = _BaseEventPDict[
    Literal['HostsTurnedOnOffEvent'],
    _HostsTurnedOnOffEventPayload,
]
type ExternalEventOccurredEventPDict = _BaseEventPDict[
    Literal['ExternalEventOccurredEvent'],
    _ExternalEventOccurredEventPayload,
]


type EventPDict = (
    JobSubmittedEventPDict
    | JobCompletedEventPDict
    | RejectJobEventPDict
    | ExecuteJobEventPDict
    | KillJobsEventPDict
    | JobsKilledEventPDict
    | RegisterProfileEventPDict
    | RegisterJobEventPDict
    | CallMeLaterEventPDict
    | RequestedCallEventPDict
    | StopCallMeLaterEventPDict
    | EDCHelloEventPDict
    | SimulationBeginsEventPDict
    | SimulationEndsEventPDict
    | AllStaticJobsHaveBeenSubmittedEventPDict
    | AllStaticExternalEventsHaveBeenInjectedEventPDict
    | FinishRegistrationEventPDict
    | ForceSimulationStopEventPDict
    | ChangeHostsPStateEventPDict
    | HostsPStateChangedEventPDict
    | TurnOnOffHostsEventPDict
    | HostsTurnedOnOffEventPDict
    | ExternalEventOccurredEventPDict
)


# ----- message wire format -----


class MessagePDict(TypedDict):
    now: Timestamp
    events: list[EventPDict]


# ----- host mappings wire format -----


class _BaseMappingStrategyPDict[T: LiteralString, P](TypedDict):
    executor_placement_type: T
    executor_placement: P


class _PredefinedMappingPayload(TypedDict):
    strategy: Literal[
        'SpreadOverHostsFirst',
        'FillOneHostCoresFirst',
    ]


class _CustomMappingPayload(TypedDict):
    mapping: list[int]


type PredefinedMappingPDict = _BaseMappingStrategyPDict[
    Literal['PredefinedExecutorPlacementStrategyWrapper'],
    _PredefinedMappingPayload,
]
type CustomMappingPDict = _BaseMappingStrategyPDict[
    Literal['CustomExecutorToHostMapping'],
    _CustomMappingPayload,
]
type MappingStrategyPDict = PredefinedMappingPDict | CustomMappingPDict


# It would be nice if we could automatically define JobAllocationPDict and
# ProfileAllocationPDict variant compatible with each corresponding
# MappingStrategyPDict variant.
# Here, we only check the literal used: this is not ideal, but this should let
# the type checker catch most incoherences.


class _BaseJobAllocationPDict[T: LiteralString, P](_BaseMappingStrategyPDict[T, P]):
    host_allocation: str


class _BaseProfileAllocationPDict[T: LiteralString, P](_BaseJobAllocationPDict[T, P]):
    profile_id: ProfileId


type JobAllocationPDict = _BaseJobAllocationPDict[
    Literal[
        'PredefinedExecutorPlacementStrategyWrapper',
        'CustomExecutorToHostMapping',
    ],
    Any,
]

type ProfileAllocationPDict = _BaseJobAllocationPDict[
    Literal[
        'PredefinedExecutorPlacementStrategyWrapper',
        'CustomExecutorToHostMapping',
    ],
    Any,
]


class StorageHostMappingPDict(TypedDict):
    storage_name: StorageId
    host_id: HostId


# ----- jobs and profiles wire format -----


class JobPDict(TypedDict):
    profile_id: ProfileId
    resource_request: int
    walltime: NotRequired[Duration]
    extra_data: NotRequired[str]


# The Profile union of the batprotocol is never directly used: it is always
# wrapped in the ProfileAndId table.
# Hence, the _BaseProfilePDict describes the structure of the ProfileAndId
# table.
class _BaseProfilePDict[T: LiteralString, P](TypedDict):
    id: ProfileId
    profile_type: T
    profile: P
    extra_data: NotRequired[str]


type _HomogeneousProfileGenerationStrategyLiteral = Literal[
    'DefinedAmountsUsedForEachValue',
    'DefinedAmountsSpreadUniformly',
]


class _DelayProfilePayload(TypedDict):
    delay: Duration


class _ParallelTaskProfilePayload(TypedDict):
    computation_vector: NotRequired[list[float]]
    communication_matrix: NotRequired[list[float]]


class _SequentialCompositionProfilePayload(TypedDict):
    repetition_count: int
    profile_ids: list[ProfileId]


class _ForkJoinCompositionProfilePayload(TypedDict):
    profile_ids: list[ProfileId]


class _ParallelTaskMergeCompositionProfilePayload(TypedDict):
    profile_ids: list[ProfileId]


class _ParallelTaskHomogeneousProfilePayload(TypedDict):
    computation_amount: float
    communication_amount: float
    generation_strategy: _HomogeneousProfileGenerationStrategyLiteral


class _ParallelTaskOnStorageHomogeneousProfilePayload(TypedDict):
    storage_name: StorageId
    bytes_to_read: float
    bytes_to_write: float
    generation_strategy: _HomogeneousProfileGenerationStrategyLiteral


class _ParallelTaskDataStagingBetweenStoragesProfilePayload(TypedDict):
    bytes_to_transfer: float
    emitter_storage_name: StorageId
    receiver_storage_name: StorageId


class _TraceReplayProfilePayload(TypedDict):
    trace_type: Literal[
        'SMPI',
        'FractionalComputation',
    ]
    filename: str


type DelayProfilePDict = _BaseProfilePDict[
    Literal['DelayProfile'],
    _DelayProfilePayload,
]
type ParallelTaskProfilePDict = _BaseProfilePDict[
    Literal['ParallelTaskProfile'],
    _ParallelTaskProfilePayload,
]
type SequentialCompositionProfilePDict = _BaseProfilePDict[
    Literal['SequentialCompositionProfile'],
    _SequentialCompositionProfilePayload,
]
type ForkJoinCompositionProfilePDict = _BaseProfilePDict[
    Literal['ForkJoinCompositionProfile'],
    _ForkJoinCompositionProfilePayload,
]
type ParallelTaskMergeCompositionProfilePDict = _BaseProfilePDict[
    Literal['ParallelTaskMergeCompositionProfile'],
    _ParallelTaskMergeCompositionProfilePayload,
]
type ParallelTaskHomogeneousProfilePDict = _BaseProfilePDict[
    Literal['ParallelTaskHomogeneousProfile'],
    _ParallelTaskHomogeneousProfilePayload,
]
type ParallelTaskOnStorageHomogeneousProfilePDict = _BaseProfilePDict[
    Literal['ParallelTaskOnStorageHomogeneousProfile'],
    _ParallelTaskOnStorageHomogeneousProfilePayload,
]
type ParallelTaskDataStagingBetweenStoragesProfilePDict = _BaseProfilePDict[
    Literal['ParallelTaskDataStagingBetweenStoragesProfile'],
    _ParallelTaskDataStagingBetweenStoragesProfilePayload,
]
type TraceReplayProfilePDict = _BaseProfilePDict[
    Literal['TraceReplayProfile'],
    _TraceReplayProfilePayload,
]


type ProfilePDict = (
    DelayProfilePDict
    | ParallelTaskProfilePDict
    | SequentialCompositionProfilePDict
    | ForkJoinCompositionProfilePDict
    | ParallelTaskMergeCompositionProfilePDict
    | ParallelTaskHomogeneousProfilePDict
    | ParallelTaskOnStorageHomogeneousProfilePDict
    | ParallelTaskDataStagingBetweenStoragesProfilePDict
    | TraceReplayProfilePDict
)



import importlib.util
import sys
from pathlib import Path

from pybatsim.batsim import Batsim, ExternalDecisionComponent


def _load_edc_from_path(location: Path, name: str) -> type[ExternalDecisionComponent]:
    """Load the EDC class 'name' from filepath 'location'."""

    # load filepath as a module
    # see https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    spec = importlib.util.spec_from_file_location(
        '__pybatsim_prototype_edc_pseudo_module', location
    )
    module = importlib.util.module_from_spec(spec)
    sys.modules['__pybatsim_prototype_edc_pseudo_module'] = module
    spec.loader.exec_module(module)

    # retrieve EDC class in loaded module
    return getattr(module, name)


class PrototypeEDC(ExternalDecisionComponent):
    """
    Load an unpackaged External Decision Component from a file.

    This convenience class proxies an EDC from a file.
    It is meant for prototyping, and we strongly advise again using it!
    **Do prefer a proper package declaring pybatsim as a dependency.**

    An example package is provided in the official repository
    (https://gitlab.inria.fr/batsim/pybatsim/) in the `pybatsim-example`
    directory.

    The scheduler expects to find in the options dict the parameters:
        - EDC_file (string): the filepath (location) containing the external
          decision component;
        - EDC_class (string): the name of the class to load.

    The user is responsible for launching pybatsim in an environment where all
    dependencies are available.
    The class MUST implement the ExternalDecisionComponent protocol.
    """

    _proxied_edc: ExternalDecisionComponent

    def __init__(self, batsim: Batsim, options):
        edc_location: Path = Path(options['EDC_file'])
        edc_name: str = options['EDC_class']

        edc_cls = _load_edc_from_path(edc_location, edc_name)
        self._proxied_edc = edc_cls(batsim, options)

    # forward **all** calls to the proxied EDC
    def __getattribute__(self, name):
        edc = object.__getattribute__(self, '_proxied_edc')  # avoids infinite recursion
        return getattr(edc, name)



"""
Workload-related data structures.
"""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import Any, Final, Self, final, override

from procset import ProcSet

from .batprotocol import (
    CustomMappingPDict,
    DelayProfilePDict,
    Duration,
    ForkJoinCompositionProfilePDict,
    JobId,
    JobPDict,
    MappingStrategyPDict,
    ParallelTaskDataStagingBetweenStoragesProfilePDict,
    ParallelTaskHomogeneousProfilePDict,
    ParallelTaskMergeCompositionProfilePDict,
    ParallelTaskOnStorageHomogeneousProfilePDict,
    ParallelTaskProfilePDict,
    PredefinedMappingPDict,
    ProfileId,
    ProfilePDict,
    SequentialCompositionProfilePDict,
    StorageId,
    Timestamp,
    TraceReplayProfilePDict,
)


class Job:
    # attributes defined at initialization
    id_: JobId
    resource_request: int
    walltime: Duration | None
    profile_id: ProfileId
    extra_data: str | None

    # attributes defined later
    submission_time: Timestamp | None = None  # set by Batsim
    allocation: ProcSet | None = None  # set by the EDC

    def __init__(
        self,
        id_: JobId,
        resource_request: int,
        profile_id: ProfileId,
        walltime: Duration | None = None,
        extra_data: str | None = None,
    ):
        self.id_ = id_
        self.resource_request = resource_request
        self.walltime = walltime
        self.profile_id = profile_id
        self.extra_data = extra_data

    def to_protocol_dict(self) -> JobPDict:
        protocol_dict: JobPDict = {
            'resource_request': self.resource_request,
            'profile_id': self.profile_id,
        }
        if self.walltime is not None:
            protocol_dict['walltime'] = self.walltime
        if self.extra_data is not None:
            protocol_dict['extra_data'] = self.extra_data
        return protocol_dict


class PlacementPolicy:
    mapping_strategy: MappingStrategy
    # TODO: integrate profile_allocation_override and storage_placement

    def __init__(self, mapping_strategy: MappingStrategy):
        self.mapping_strategy = mapping_strategy

    @classmethod
    def default(cls) -> Self:
        """Return the default placement policy."""
        return cls(mapping_strategy=MappingStrategy.default())


# pybatsim representation of ExecutorPlacement
class MappingStrategy:
    @abstractmethod
    def to_protocol_dict(self) -> MappingStrategyPDict: ...

    @staticmethod
    def default() -> MappingStrategy:
        """Return the default mapping strategy."""
        return SpreadHostsMapping()


class SpreadHostsMapping(MappingStrategy):
    @override
    def to_protocol_dict(self) -> PredefinedMappingPDict:
        return {
            'executor_placement_type': 'PredefinedExecutorPlacementStrategyWrapper',
            'executor_placement': {
                'strategy': 'SpreadOverHostsFirst',
            },
        }


class FillHostsMapping(MappingStrategy):
    @override
    def to_protocol_dict(self) -> PredefinedMappingPDict:
        return {
            'executor_placement_type': 'PredefinedExecutorPlacementStrategyWrapper',
            'executor_placement': {
                'strategy': 'FillOneHostCoresFirst',
            },
        }


class CustomMapping(MappingStrategy):
    mapping: list[int]

    def __init__(self, mapping: list[int]):
        self.mapping = mapping

    @override
    def to_protocol_dict(self) -> CustomMappingPDict:
        return {
            'executor_placement_type': 'CustomExecutorToHostMapping',
            'executor_placement': {
                'mapping': self.mapping,
            },
        }


# class KillProgress
# TODO: implement me


class FinalState(Enum):
    SUCCESS = 'COMPLETED_SUCCESSFULLY'
    FAILED = 'COMPLETED_FAILED'
    WALLTIME_REACHED = 'COMPLETED_WALLTIME_REACHED'
    KILLED = 'COMPLETED_KILLED'
    REJECTED = 'REJECTED'


class Profile:
    _batprotocol_profiles: Final[dict[str, type[Profile]]] = {}

    id_: ProfileId
    extra_data: str | None

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        cls._batprotocol_profiles[cls.__name__] = cls

    def __init__(self, id_: ProfileId, extra_data: str | None = None):
        self.id_ = id_
        self.extra_data = extra_data

    # Serialization and deserialization are made against ProfileAndId structure.

    @abstractmethod
    def to_protocol_dict(self) -> ProfilePDict: ...

    # Ideally, we would like to only have a from_protocol_dict() method, and no
    # _deserialize_protocol_dict() method.
    # This is not achievable via overriding in the concrete subclasses as this
    # would violate the Liskov substitution principle: method arguments are
    # contravariant, and we want to further specify the type in our case.
    # _deserialize_protocol_dict() does not uphold the Liskov substitution
    # principle either but this is not exposed to user-facing API.

    @final
    @classmethod
    def from_protocol_dict(cls, payload: ProfilePDict) -> Profile:
        concrete_cls = cls._batprotocol_profiles[payload['profile_type']]
        return concrete_cls._deserialize_protocol_dict(payload)  # noqa: SLF001 (reason: private in module scope)

    @classmethod
    @abstractmethod
    def _deserialize_protocol_dict(cls, payload) -> Self: ...  # type: ignore[no-untyped-def]


class DelayProfile(Profile):
    delay: Duration

    @override
    def __init__(self, id_: ProfileId, delay: Duration, extra_data: str | None = None):
        super().__init__(id_, extra_data)
        self.delay = delay

    @override
    def to_protocol_dict(self) -> DelayProfilePDict:
        protocol_dict: DelayProfilePDict = {
            'id': self.id_,
            'profile_type': 'DelayProfile',
            'profile': {
                'delay': self.delay,
            },
        }
        if self.extra_data is not None:
            protocol_dict['extra_data'] = self.extra_data
        return protocol_dict

    @override
    @classmethod
    def _deserialize_protocol_dict(cls, payload: DelayProfilePDict) -> Self:
        return cls(
            id_=payload['id'],
            delay=payload['profile']['delay'],
            extra_data=payload.get('extra_data'),
        )


class ParallelTaskProfile(Profile):
    computation_vector: list[float] | None
    communication_matrix: list[float] | None

    @override
    def __init__(
        self,
        id_: ProfileId,
        computation_vector: list[float] | None,
        communication_matrix: list[float] | None,
        extra_data: str | None = None,
    ):
        super().__init__(id_, extra_data)
        self.computation_vector = computation_vector
        self.communication_matrix = communication_matrix

    @override
    def to_protocol_dict(self) -> ParallelTaskProfilePDict:
        protocol_dict: ParallelTaskProfilePDict = {
            'id': self.id_,
            'profile_type': 'ParallelTaskProfile',
            'profile': {},
        }
        if self.computation_vector is not None:
            protocol_dict['profile']['computation_vector'] = self.computation_vector
        if self.communication_matrix is not None:
            protocol_dict['profile']['communication_matrix'] = self.communication_matrix
        if self.extra_data is not None:
            protocol_dict['extra_data'] = self.extra_data
        return protocol_dict

    @override
    @classmethod
    def _deserialize_protocol_dict(cls, payload: ParallelTaskProfilePDict) -> Self:
        return cls(
            id_=payload['id'],
            computation_vector=payload['profile'].get('computation_vector'),
            communication_matrix=payload['profile'].get('communication_matrix'),
            extra_data=payload.get('extra_data'),
        )


class SequentialCompositionProfile(Profile):
    repetition_count: int = 1
    profile_ids: list[ProfileId]

    @override
    def __init__(
        self,
        id_: ProfileId,
        profile_ids: list[ProfileId],
        repetition_count: int = 1,
        extra_data: str | None = None,
    ):
        super().__init__(id_, extra_data)
        self.repetition_count = repetition_count
        self.profile_ids = profile_ids

    @override
    def to_protocol_dict(self) -> SequentialCompositionProfilePDict:
        protocol_dict: SequentialCompositionProfilePDict = {
            'id': self.id_,
            'profile_type': 'SequentialCompositionProfile',
            'profile': {
                'repetition_count': 1,
                'profile_ids': self.profile_ids,
            },
        }
        if self.extra_data is not None:
            protocol_dict['extra_data'] = self.extra_data
        return protocol_dict

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: SequentialCompositionProfilePDict,
    ) -> Self:
        return cls(
            id_=payload['id'],
            repetition_count=payload['profile']['repetition_count'],
            profile_ids=payload['profile']['profile_ids'],
            extra_data=payload.get('extra_data'),
        )


class ForkJoinCompositionProfile(Profile):
    profile_ids: list[ProfileId]

    @override
    def __init__(
        self,
        id_: ProfileId,
        profile_ids: list[ProfileId],
        extra_data: str | None = None,
    ):
        super().__init__(id_, extra_data)
        self.profile_ids = profile_ids

    @override
    def to_protocol_dict(self) -> ForkJoinCompositionProfilePDict:
        protocol_dict: ForkJoinCompositionProfilePDict = {
            'id': self.id_,
            'profile_type': 'ForkJoinCompositionProfile',
            'profile': {
                'profile_ids': self.profile_ids,
            },
        }
        if self.extra_data is not None:
            protocol_dict['extra_data'] = self.extra_data
        return protocol_dict

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: ForkJoinCompositionProfilePDict,
    ) -> Self:
        return cls(
            id_=payload['id'],
            profile_ids=payload['profile']['profile_ids'],
            extra_data=payload.get('extra_data'),
        )


class ParallelTaskMergeCompositionProfile(Profile):
    profile_ids: list[ProfileId]

    @override
    def __init__(
        self,
        id_: ProfileId,
        profile_ids: list[ProfileId],
        extra_data: str | None = None,
    ):
        super().__init__(id_, extra_data)
        self.profile_ids = profile_ids

    @override
    def to_protocol_dict(self) -> ParallelTaskMergeCompositionProfilePDict:
        protocol_dict: ParallelTaskMergeCompositionProfilePDict = {
            'id': self.id_,
            'profile_type': 'ParallelTaskMergeCompositionProfile',
            'profile': {
                'profile_ids': self.profile_ids,
            },
        }
        if self.extra_data is not None:
            protocol_dict['extra_data'] = self.extra_data
        return protocol_dict

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: ParallelTaskMergeCompositionProfilePDict,
    ) -> Self:
        return cls(
            id_=payload['id'],
            profile_ids=payload['profile']['profile_ids'],
            extra_data=payload.get('extra_data'),
        )


class HomogeneousProfileGenerationStrategy(Enum):
    AMOUNT_UNIT = 'DefinedAmountsUsedForEachValue'
    AMOUNT_GLOBAL = 'DefinedAmountsSpreadUniformly'


class ParallelTaskHomogeneousProfile(Profile):
    computation_amount: float
    communication_amount: float
    generation_strategy: HomogeneousProfileGenerationStrategy

    @override
    def __init__(
        self,
        id_: ProfileId,
        computation_amount: float,
        communication_amount: float,
        generation_strategy: HomogeneousProfileGenerationStrategy,
        extra_data: str | None = None,
    ):
        super().__init__(id_, extra_data)
        self.computation_amount = computation_amount
        self.communication_amount = communication_amount
        self.generation_strategy = generation_strategy

    @override
    def to_protocol_dict(self) -> ParallelTaskHomogeneousProfilePDict:
        protocol_dict: ParallelTaskHomogeneousProfilePDict = {
            'id': self.id_,
            'profile_type': 'ParallelTaskHomogeneousProfile',
            'profile': {
                'computation_amount': self.computation_amount,
                'communication_amount': self.communication_amount,
                'generation_strategy': self.generation_strategy.value,
            },
        }
        if self.extra_data is not None:
            protocol_dict['extra_data'] = self.extra_data
        return protocol_dict

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: ParallelTaskHomogeneousProfilePDict,
    ) -> Self:
        return cls(
            id_=payload['id'],
            computation_amount=payload['profile'].get('computation_amount'),
            communication_amount=payload['profile'].get('communication_amount'),
            generation_strategy=HomogeneousProfileGenerationStrategy(
                payload['profile']['generation_strategy']
            ),
            extra_data=payload.get('extra_data'),
        )


class ParallelTaskOnStorageHomogeneousProfile(Profile):
    storage_name: StorageId
    bytes_to_read: float
    bytes_to_write: float
    generation_strategy: HomogeneousProfileGenerationStrategy

    @override
    def __init__(
        self,
        id_: ProfileId,
        storage_name: StorageId,
        bytes_to_read: float,
        bytes_to_write: float,
        generation_strategy: HomogeneousProfileGenerationStrategy,
        extra_data: str | None = None,
    ):
        super().__init__(id_, extra_data)
        self.storage_name = storage_name
        self.bytes_to_read = bytes_to_read
        self.bytes_to_write = bytes_to_write
        self.generation_strategy = generation_strategy

    @override
    def to_protocol_dict(self) -> ParallelTaskOnStorageHomogeneousProfilePDict:
        protocol_dict: ParallelTaskOnStorageHomogeneousProfilePDict = {
            'id': self.id_,
            'profile_type': 'ParallelTaskOnStorageHomogeneousProfile',
            'profile': {
                'storage_name': self.storage_name,
                'bytes_to_read': self.bytes_to_read,
                'bytes_to_write': self.bytes_to_write,
                'generation_strategy': self.generation_strategy.value,
            },
        }
        if self.extra_data is not None:
            protocol_dict['extra_data'] = self.extra_data
        return protocol_dict

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: ParallelTaskOnStorageHomogeneousProfilePDict,
    ) -> Self:
        return cls(
            id_=payload['id'],
            storage_name=payload['profile']['storage_name'],
            bytes_to_read=payload['profile']['bytes_to_read'],
            bytes_to_write=payload['profile']['bytes_to_write'],
            generation_strategy=HomogeneousProfileGenerationStrategy(
                payload['profile']['generation_strategy']
            ),
            extra_data=payload.get('extra_data'),
        )


class ParallelTaskDataStagingBetweenStoragesProfile(Profile):
    bytes_to_transfer: float
    emitter_storage_name: StorageId
    receiver_storage_name: StorageId

    @override
    def __init__(
        self,
        id_: ProfileId,
        bytes_to_transfer: float,
        emitter_storage_name: StorageId,
        receiver_storage_name: StorageId,
        extra_data: str | None = None,
    ):
        super().__init__(id_, extra_data)
        self.bytes_to_transfer = bytes_to_transfer
        self.emitter_storage_name = emitter_storage_name
        self.receiver_storage_name = receiver_storage_name

    @override
    def to_protocol_dict(self) -> ParallelTaskDataStagingBetweenStoragesProfilePDict:
        protocol_dict: ParallelTaskDataStagingBetweenStoragesProfilePDict = {
            'id': self.id_,
            'profile_type': 'ParallelTaskDataStagingBetweenStoragesProfile',
            'profile': {
                'bytes_to_transfer': self.bytes_to_transfer,
                'emitter_storage_name': self.emitter_storage_name,
                'receiver_storage_name': self.receiver_storage_name,
            },
        }
        if self.extra_data is not None:
            protocol_dict['extra_data'] = self.extra_data
        return protocol_dict

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: ParallelTaskDataStagingBetweenStoragesProfilePDict,
    ) -> Self:
        return cls(
            id_=payload['id'],
            bytes_to_transfer=payload['profile']['bytes_to_transfer'],
            emitter_storage_name=payload['profile']['emitter_storage_name'],
            receiver_storage_name=payload['profile']['receiver_storage_name'],
            extra_data=payload.get('extra_data'),
        )


class TraceType(Enum):
    SMPI = 'SMPI'
    FRACTIONAL_COMPUTATION = 'FractionalComputation'


class TraceReplayProfile(Profile):
    trace_type: TraceType
    filename: str

    @override
    def __init__(
        self,
        id_: ProfileId,
        trace_type: TraceType,
        filename: str,
        extra_data: str | None = None,
    ):
        super().__init__(id_, extra_data)
        self.trace_type = trace_type
        self.filename = filename

    @override
    def to_protocol_dict(self) -> TraceReplayProfilePDict:
        protocol_dict: TraceReplayProfilePDict = {
            'id': self.id_,
            'profile_type': 'TraceReplayProfile',
            'profile': {
                'trace_type': self.trace_type.value,
                'filename': self.filename,
            },
        }
        if self.extra_data is not None:
            protocol_dict['extra_data'] = self.extra_data
        return protocol_dict

    @override
    @classmethod
    def _deserialize_protocol_dict(cls, payload: TraceReplayProfilePDict) -> Self:
        return cls(
            id_=payload['id'],
            trace_type=TraceType(payload['profile']['trace_type']),
            filename=payload['profile']['filename'],
            extra_data=payload.get('extra_data'),
        )



"""
Events of the batprotocol.

The following batprotocol v1.0.0 events have an experimental implementation.
They may change behavior in a breaking manner upon minor version update.

- JobsKilledEvent
- SimulationBeginsEvent


The following batprotocol v1.0.0 events are unsupported.
They may be implemented in a subsequent minor version.

- CreateProbeEvent
- StopProbeEvent
- TriggerProbeEvent
- ResetProbeEvent
- ProbeDataEmittedEvent
"""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import Any, Final, Self, final, override

from procset import ProcSet

from .batprotocol import (
    AllStaticExternalEventsHaveBeenInjectedEventPDict,
    AllStaticJobsHaveBeenSubmittedEventPDict,
    CallId,
    CallMeLaterEventPDict,
    ChangeHostsPStateEventPDict,
    EDCHelloEventPDict,
    EventPDict,
    ExecuteJobEventPDict,
    ExternalEventOccurredEventPDict,
    FinishRegistrationEventPDict,
    ForceSimulationStopEventPDict,
    HostsPStateChangedEventPDict,
    HostsTurnedOnOffEventPDict,
    JobCompletedEventPDict,
    JobId,
    JobsKilledEventPDict,
    JobSubmittedEventPDict,
    KillJobsEventPDict,
    RegisterJobEventPDict,
    RegisterProfileEventPDict,
    RejectJobEventPDict,
    RequestedCallEventPDict,
    SimulationBeginsEventPDict,
    SimulationEndsEventPDict,
    StopCallMeLaterEventPDict,
    Timestamp,
    TurnOnOffHostsEventPDict,
    _SimulationBeginsEventPayload,
)
from .core import SimulationMetadata
from .job import FinalState, Job, PlacementPolicy, Profile


def __experimental(*, reason: str | None = None):  # type: ignore[no-untyped-def]
    """Decorator to identify experimental features."""

    def decorator(obj):  # type: ignore[no-untyped-def]
        # Craft experimental note.
        entity_type = 'class' if isinstance(obj, type) else 'function'
        note = f'This {entity_type} is experimental and is subject to backward-incompatible changes upon minor releases.'  # noqa: E501 (reason: grepability)
        if reason:
            note += f'\nReason: {reason}'

        # Update docstring.
        if obj.__doc__:
            obj.__doc__ += f'\n\n{note}'
        else:
            obj.__doc__ = note

        return obj

    return decorator


# abstract base classes  -------------------------------------------------------


class Event:
    timestamp: Timestamp

    def __new__(cls, *_args: Any, **_kwargs: Any) -> Self:
        # Ensure Event, RxEvent or TxEvent are not instantiated, see:
        # - https://stackoverflow.com/a/7990308
        # - https://docs.python.org/3/reference/datamodel.html#object.__new__
        if cls in (Event, RxEvent, TxEvent):
            err_msg = f'Cannot instantiate abstract class {cls.__name__}'
            raise TypeError(err_msg)

        # We pass only cls to object.__new__() as we also define __init__().
        # However, we need to keep args and kwargs in the signature as they are
        # forwarded to __init__().
        return object.__new__(cls)

    def __init__(self, timestamp: Timestamp):
        self.timestamp = timestamp


class RxEvent(Event):
    _batprotocol_rx_events: Final[dict[str, type[RxEvent]]] = {}

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        cls._batprotocol_rx_events[cls.__name__] = cls

    # Ideally, we would like to only have a from_protocol_dict() method, and no
    # _deserialize_protocol_dict() method.
    # This is not achievable via overriding in the concrete subclasses as this
    # would violate the Liskov substitution principle: method arguments are
    # contravariant, and we want to further specify the type in our case.
    # _deserialize_protocol_dict() does not uphold the Liskov substitution
    # principle either but this is not exposed to user-facing API.

    @final
    @classmethod
    def from_protocol_dict(cls, payload: EventPDict) -> RxEvent:
        concrete_cls = cls._batprotocol_rx_events[payload['event_type']]
        return concrete_cls._deserialize_protocol_dict(payload)  # noqa: SLF001 (reason: private in module scope)

    @classmethod
    @abstractmethod
    def _deserialize_protocol_dict(cls, payload) -> Self: ...  # type: ignore[no-untyped-def]


class TxEvent(Event):
    @abstractmethod
    def to_protocol_dict(self) -> EventPDict: ...


# concrete classes  ------------------------------------------------------------


# TODO(rb):
#   dispatch elements of _payload to attributes with semantic types:
#     - SimulationMetadata
#     - (to create/define) Simulation(Args|Params|Context)
@__experimental(reason='_payload is yet to be mapped to semantic types.')
class SimulationBeginsEvent(RxEvent):
    _payload: _SimulationBeginsEventPayload  # XXX: replace with semantic attributes

    @override
    def __init__(self, timestamp: Timestamp, payload: _SimulationBeginsEventPayload):
        super().__init__(timestamp)
        self._payload = payload

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: SimulationBeginsEventPDict,
    ) -> SimulationBeginsEvent:
        return cls(
            timestamp=payload['timestamp'],
            payload=payload['event'],
        )

    @property
    def computation_host_number(self) -> int:
        return self._payload['computation_host_number']


class SimulationEndsEvent(RxEvent):
    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: SimulationEndsEventPDict,
    ) -> SimulationEndsEvent:
        return cls(
            timestamp=payload['timestamp'],
        )


class JobSubmittedEvent(RxEvent):
    job: Job
    fwd_profile: Profile | None

    @override
    def __init__(
        self,
        timestamp: Timestamp,
        job: Job,
        fwd_profile: Profile | None = None,
    ):
        super().__init__(timestamp)
        self.job = job
        self.fwd_profile = fwd_profile

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: JobSubmittedEventPDict,
    ) -> JobSubmittedEvent:
        fwd_profile_dict = payload['event'].get('profile')
        fwd_profile = (
            Profile.from_protocol_dict(fwd_profile_dict)
            if fwd_profile_dict is not None
            else None
        )

        job = Job(
            id_=payload['event']['job_id'],
            resource_request=payload['event']['job']['resource_request'],
            walltime=payload['event']['job'].get('walltime'),
            profile_id=payload['event']['job']['profile_id'],
            extra_data=payload['event']['job'].get('extra_data'),
        )
        job.submission_time = payload['event']['submission_time']

        assert fwd_profile is None or fwd_profile.id_ == job.profile_id

        return cls(
            timestamp=payload['timestamp'],
            job=job,
            fwd_profile=fwd_profile,
        )


class JobCompletedEvent(RxEvent):
    job: Job
    state: FinalState
    return_code: int

    @override
    def __init__(
        self,
        timestamp: Timestamp,
        job: Job,
        state: FinalState,
        return_code: int,
    ):
        super().__init__(timestamp)
        self.job = job
        self.state = state
        self.return_code = return_code

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: JobCompletedEventPDict,
    ) -> JobCompletedEvent:
        assert payload['__pybatsim_job'].id_ == payload['event']['job_id']
        return cls(
            timestamp=payload['timestamp'],
            job=payload['__pybatsim_job'],  # injected by deserialisation
            state=FinalState(payload['event']['state']),
            return_code=payload['event']['return_code'],
        )


@__experimental(reason='progress is yet to be mapped to a semantic type.')
class JobsKilledEvent(RxEvent):
    # TODO(rb): consider using a list of tuples
    jobs: list[Job]
    progresses: dict[JobId, dict[Any, Any]]

    @override
    def __init__(
        self,
        timestamp: Timestamp,
        jobs: list[Job],
        progresses: dict[JobId, dict[Any, Any]],
    ):
        super().__init__(timestamp)
        self.jobs = jobs
        self.progresses = progresses

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: JobsKilledEventPDict,
    ) -> JobsKilledEvent:
        progresses: dict[JobId, Any] = {}  # XXX: replace with semantic type
        for progress_dict in payload['event']['progresses']:
            job_id = progress_dict['job_id']
            # TODO: correctly deserialise the wrapped progress in a KillProgress object
            progresses[job_id] = progress_dict['wrapper']

        return cls(
            timestamp=payload['timestamp'],
            jobs=payload['__pybatsim_dead_jobs'],  # injected by deserialisation
            progresses=progresses,
        )


class RequestedCallEvent(RxEvent):
    call_id: CallId
    is_last_call: bool

    @override
    def __init__(self, timestamp: Timestamp, call_id: CallId, is_last_call: bool):
        super().__init__(timestamp)
        self.call_id = call_id
        self.is_last_call = is_last_call

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: RequestedCallEventPDict,
    ) -> RequestedCallEvent:
        return cls(
            timestamp=payload['timestamp'],
            call_id=payload['event']['call_me_later_id'],
            is_last_call=payload['event']['last_periodic_call'],
        )


class ExternalEventOccurredEvent(RxEvent):
    external_event_id: str
    # TODO: consider using an enum, might be overkill as there exists a single
    # type as of now (GenericExternalEvent)
    external_event_type: str
    external_event: dict[Any, Any]  # # XXX: replace with semantic type

    @override
    def __init__(
        self,
        timestamp: Timestamp,
        external_event_id: str,
        external_event_type: str,
        external_event: dict[Any, Any],
    ):
        super().__init__(timestamp)
        self.external_event_id = external_event_id
        self.external_event_type = external_event_type
        self.external_event = external_event

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: ExternalEventOccurredEventPDict,
    ) -> ExternalEventOccurredEvent:
        return cls(
            timestamp=payload['timestamp'],
            external_event_id=payload['event']['id'],
            external_event_type=payload['event']['external_event_type'],
            external_event=payload['event']['external_event'],
        )


class AllStaticJobsHaveBeenSubmittedEvent(RxEvent):
    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: AllStaticJobsHaveBeenSubmittedEventPDict,
    ) -> AllStaticJobsHaveBeenSubmittedEvent:
        return cls(
            timestamp=payload['timestamp'],
        )


class AllStaticExternalEventsHaveBeenInjectedEvent(RxEvent):
    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: AllStaticExternalEventsHaveBeenInjectedEventPDict,
    ) -> AllStaticExternalEventsHaveBeenInjectedEvent:
        return cls(
            timestamp=payload['timestamp'],
        )


class HostsPStateChangedEvent(RxEvent):
    host_ids: ProcSet
    pstate: int

    @override
    def __init__(self, timestamp: Timestamp, host_ids: ProcSet, pstate: int):
        super().__init__(timestamp)
        self.host_ids = host_ids
        self.pstate = pstate

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: HostsPStateChangedEventPDict,
    ) -> HostsPStateChangedEvent:
        return cls(
            timestamp=payload['timestamp'],
            host_ids=ProcSet.from_str(payload['event']['host_ids']),
            pstate=payload['event']['pstate'],
        )


class HostsTurnedOnOffEvent(RxEvent):
    host_ids: ProcSet
    state: int

    @override
    def __init__(self, timestamp: Timestamp, host_ids: ProcSet, state: int):
        super().__init__(timestamp)
        self.host_ids = host_ids
        self.state = state

    @override
    @classmethod
    def _deserialize_protocol_dict(
        cls,
        payload: HostsTurnedOnOffEventPDict,
    ) -> HostsTurnedOnOffEvent:
        return cls(
            timestamp=payload['timestamp'],
            host_ids=ProcSet.from_str(payload['event']['host_ids']),
            state=payload['event']['state'],
        )


class EDCHelloEvent(TxEvent):
    simulation_metadata: SimulationMetadata
    # TODO: consider integrating edc_* in simulation_metadata
    edc_name: str
    edc_version: str
    edc_commit: str

    @override
    def __init__(
        self,
        timestamp: Timestamp,
        simulation_metadata: SimulationMetadata,
        edc_name: str,
        edc_version: str,
        edc_commit: str,
    ):
        super().__init__(timestamp)
        self.simulation_metadata = simulation_metadata
        self.edc_name = edc_name
        self.edc_version = edc_version
        self.edc_commit = edc_commit

    @override
    def to_protocol_dict(self) -> EDCHelloEventPDict:
        return {
            'timestamp': self.timestamp,
            'event_type': 'EDCHelloEvent',
            'event': {
                **self.simulation_metadata.to_protocol_dict(),
                'decision_component_name': self.edc_name,
                'decision_component_version': self.edc_version,
                'decision_component_commit': self.edc_commit,
            },
        }


class RejectJobEvent(TxEvent):
    job: Job

    @override
    def __init__(self, timestamp: Timestamp, job: Job):
        super().__init__(timestamp)
        self.job = job

    @override
    def to_protocol_dict(self) -> RejectJobEventPDict:
        return {
            'timestamp': self.timestamp,
            'event_type': 'RejectJobEvent',
            'event': {
                'job_id': self.job.id_,
            },
        }


class ExecuteJobEvent(TxEvent):
    job: Job
    placement_policy: PlacementPolicy

    @override
    def __init__(
        self,
        timestamp: Timestamp,
        job: Job,
        placement_policy: PlacementPolicy | None = None,
    ):
        super().__init__(timestamp)
        self.job = job
        self.placement_policy = (
            PlacementPolicy.default() if placement_policy is None else placement_policy
        )

    @override
    def to_protocol_dict(self) -> ExecuteJobEventPDict:
        return {
            'timestamp': self.timestamp,
            'event_type': 'ExecuteJobEvent',
            'event': {
                'job_id': self.job.id_,
                'allocation': {
                    **self.placement_policy.mapping_strategy.to_protocol_dict(),
                    'host_allocation': str(self.job.allocation),
                },
            },
        }


class KillJobsEvent(TxEvent):
    jobs: list[Job]

    @override
    def __init__(self, timestamp: Timestamp, jobs: list[Job]):
        super().__init__(timestamp)
        self.jobs = jobs

    @override
    def to_protocol_dict(self) -> KillJobsEventPDict:
        return {
            'timestamp': self.timestamp,
            'event_type': 'KillJobsEvent',
            'event': {
                'job_ids': [job.id_ for job in self.jobs],
            },
        }


class CallMeLaterEvent(TxEvent):
    # TODO: nesting class is not Pythonic
    class TemporalTriggerType(Enum):
        OneShot = 0
        Periodic = 1

    call_id: CallId
    # XXX: replace when_type and when by a semantic type
    when_type: TemporalTriggerType
    when: dict[Any, Any]

    @override
    def __init__(
        self,
        timestamp: Timestamp,
        call_id: CallId,
        when_type: TemporalTriggerType,
        when_dict: dict[Any, Any],
    ):
        super().__init__(timestamp)
        self.call_id = call_id
        self.when_type = when_type
        self.when = when_dict

    @override
    def to_protocol_dict(self) -> CallMeLaterEventPDict:
        return {
            'timestamp': self.timestamp,
            'event_type': 'CallMeLaterEvent',
            'event': {
                'call_me_later_id': self.call_id,
                'when_type': self.when_type.name,
                'when': self.when,
            },
        }


class StopCallMeLaterEvent(TxEvent):
    call_id: CallId

    @override
    def __init__(self, timestamp: Timestamp, call_id: CallId):
        super().__init__(timestamp)
        self.call_id = call_id

    @override
    def to_protocol_dict(self) -> StopCallMeLaterEventPDict:
        return {
            'timestamp': self.timestamp,
            'event_type': 'StopCallMeLaterEvent',
            'event': {
                'call_me_later_id': self.call_id,
            },
        }


class ForceSimulationStopEvent(TxEvent):
    @override
    def to_protocol_dict(self) -> ForceSimulationStopEventPDict:
        return {
            'timestamp': self.timestamp,
            'event_type': 'ForceSimulationStopEvent',
            'event': {},
        }


class FinishRegistrationEvent(TxEvent):
    @override
    def to_protocol_dict(self) -> FinishRegistrationEventPDict:
        return {
            'timestamp': self.timestamp,
            'event_type': 'FinishRegistrationEvent',
            'event': {},
        }


class RegisterJobEvent(TxEvent):
    job: Job

    @override
    def __init__(self, timestamp: Timestamp, job: Job):
        super().__init__(timestamp)
        self.job = job
        # XXX: job.submission_time will never be set if ack of dynamic jobs is
        # disabled: consider injecting timestamp as job.submission_time

    @override
    def to_protocol_dict(self) -> RegisterJobEventPDict:
        return {
            'timestamp': self.timestamp,
            'event_type': 'RegisterJobEvent',
            'event': {
                'job_id': self.job.id_,
                'job': self.job.to_protocol_dict(),
            },
        }


class RegisterProfileEvent(TxEvent):
    profile: Profile

    @override
    def __init__(self, timestamp: Timestamp, profile: Profile):
        super().__init__(timestamp)
        self.profile = profile

    @override
    def to_protocol_dict(self) -> RegisterProfileEventPDict:
        return {
            'timestamp': self.timestamp,
            'event_type': 'RegisterProfileEvent',
            'event': {
                'profile': self.profile.to_protocol_dict(),
            },
        }


class ChangeHostsPStateEvent(TxEvent):
    host_ids: ProcSet
    pstate: int

    @override
    def __init__(self, timestamp: Timestamp, host_ids: ProcSet, pstate: int):
        super().__init__(timestamp)
        self.host_ids = host_ids
        self.pstate = pstate

    @override
    def to_protocol_dict(self) -> ChangeHostsPStateEventPDict:
        return {
            'timestamp': self.timestamp,
            'event_type': 'ChangeHostsPStateEvent',
            'event': {
                'host_ids': str(self.host_ids),
                'pstate': self.pstate,
            },
        }


class TurnOnOffHostsEvent(TxEvent):
    host_ids: ProcSet
    state: int

    @override
    def __init__(self, timestamp: Timestamp, host_ids: ProcSet, state: int):
        super().__init__(timestamp)
        self.host_ids = host_ids
        self.state = state

    @override
    def to_protocol_dict(self) -> TurnOnOffHostsEventPDict:
        return {
            'timestamp': self.timestamp,
            'event_type': 'TurnOnOffHostsEvent',
            'event': {
                'host_ids': str(self.host_ids),
                'state': self.state,
            },
        }



from .batsim import (
    SERIALIZATION_FORMAT_JSON,
    WORKLOAD_JOB_SEPARATOR,
    Batsim,
    ExternalDecisionComponent,
    Scheduler,
)
from .core import SimulationFeatures, SimulationMetadata
from .events import (
    AllStaticExternalEventsHaveBeenInjectedEvent,
    AllStaticJobsHaveBeenSubmittedEvent,
    CallMeLaterEvent,
    ChangeHostsPStateEvent,
    EDCHelloEvent,
    Event,
    ExecuteJobEvent,
    ExternalEventOccurredEvent,
    FinishRegistrationEvent,
    ForceSimulationStopEvent,
    HostsPStateChangedEvent,
    HostsTurnedOnOffEvent,
    JobCompletedEvent,
    JobsKilledEvent,
    JobSubmittedEvent,
    KillJobsEvent,
    RegisterJobEvent,
    RegisterProfileEvent,
    RejectJobEvent,
    RequestedCallEvent,
    RxEvent,
    SimulationBeginsEvent,
    SimulationEndsEvent,
    StopCallMeLaterEvent,
    TurnOnOffHostsEvent,
    TxEvent,
)
from .job import Job, Profile

__all__ = [
    'SERIALIZATION_FORMAT_JSON',
    'WORKLOAD_JOB_SEPARATOR',
    'AllStaticExternalEventsHaveBeenInjectedEvent',
    'AllStaticJobsHaveBeenSubmittedEvent',
    'Batsim',
    'CallMeLaterEvent',
    'ChangeHostsPStateEvent',
    'EDCHelloEvent',
    'Event',
    'ExecuteJobEvent',
    'ExternalDecisionComponent',
    'ExternalEventOccurredEvent',
    'FinishRegistrationEvent',
    'ForceSimulationStopEvent',
    'HostsPStateChangedEvent',
    'HostsTurnedOnOffEvent',
    'Job',
    'JobCompletedEvent',
    'JobSubmittedEvent',
    'JobsKilledEvent',
    'KillJobsEvent',
    'Profile',
    'RegisterJobEvent',
    'RegisterProfileEvent',
    'RejectJobEvent',
    'RequestedCallEvent',
    'RxEvent',
    'Scheduler',
    'SimulationBeginsEvent',
    'SimulationEndsEvent',
    'SimulationFeatures',
    'SimulationMetadata',
    'StopCallMeLaterEvent',
    'TurnOnOffHostsEvent',
    'TxEvent',
]
